import greenfoot.*;  // (World, Actor, GreenfootImage, Greenfoot and MouseInfo)
import java.awt.Color;
import java.util.List;
import java.util.ArrayList;

/**
 * This class is the main class of the DropdownList. You can make objects of this class with many different consrtuktors.
 * To create a list with objects in it you have to give the dropdownlist a java.util.List<Element> as parameter which should be added to the list.
 * To add your own Objects to the list these objects have to implement the interface Element.
 * 
 * @author Gevater_Tod4711
 * @version 11.11.2013
 */
public class DropdownList extends Actor
{
    private boolean elementOfList;
    private boolean listDisplayed = false;
    
    private int width;
    private int height;
    
    private Color listColor;
    private Color textColor;
    
    private Element element;
    
    private GreenfootActionListener greenfootActionListener;
    
    private DropdownList callingList;
    private DropdownList emptyElement;
    
    private ArrayList<DropdownList> listElements;
    
    /**
     * Creates a new DropdownList with the given elements, a default size of 375 x 42, with black letters on white ground and without an element that is shown directly.
     * 
     * @param elements
     *      A list of elements you want to add in the list.
     *      If you want an empty list you have to give an empty java.util.List<Element> as parameter but not null.
     *      Null is not excepted as a parameter.
     * 
     * @throws
     *      IllegalArgumentException is thrown if the parameter elements is null.
     */
    public DropdownList(List<? extends Element> elements) throws IllegalArgumentException {
        this(elements, null, 375, 42, Color.white, Color.black);
    }
    /**
     * Creates a new DropdownList with the given elements, a default size of 375 x 42, with black letters on white ground and with an element that is shown directly.
     * 
     * @param elements
     *      A list of elements you want to add in the list.
     *      If you want an empty list you have to give an empty java.util.List<Element> as parameter but not null.
     *      Null is not excepted as a parameter.
     * 
     * @param firstShownElement
     *      The element that is shown directly after creating the DropdownList.
     * 
     * @throws
     *      IllegalArgumentException is thrown if the parameter elements is null.
     */
    public DropdownList(List<? extends Element> elements, Element firstShownElement) throws IllegalArgumentException {
        this(elements, firstShownElement, 375, 42, Color.white, Color.black);
    }
    /**
     * Creates a new DropdownList with the given elements, a default size of 375 x 42, with black letters on a ground which color is the given parameter and without an element that is shown directly.
     * 
     * @param elements
     *      A list of elements you want to add in the list.
     *      If you want an empty list you have to give an empty java.util.List<Element> as parameter but not null.
     *      Null is not excepted as a parameter.
     * 
     * @param listColor
     *      The background color of the list.
     * 
     * @throws
     *      IllegalArgumentException is thrown if the parameter elements is null or if the Color listColor is null.
     */
    public DropdownList(List<? extends Element> elements, Color listColor) throws IllegalArgumentException {
        this(elements, null, 375, 40, listColor, Color.black);
    }
    /**
     * Creates a new DropdownList with the given elements, a default size of 375 x 42, with black letters on a ground which color is the given parameter and with an element that is shown directly.
     * 
     * @param elements
     *      A list of elements you want to add in the list.
     *      If you want an empty list you have to give an empty java.util.List<Element> as parameter but not null.
     *      Null is not excepted as a parameter.
     * 
     * @param firstShownElement
     *      The element that is shown directly after creating the DropdownList.
     * 
     * @param listColor
     *      The background color of the list.
     * 
     * @throws
     *      IllegalArgumentException is thrown if the parameter elements is null or if the Color listColor is null.
     */
    public DropdownList(List<? extends Element> elements, Element firstShownElement, Color listColor) throws IllegalArgumentException {
        this(elements, firstShownElement, 375, 40, listColor, Color.black);
    }
    /**
     * Creates a new DropdownList with the given elements, a default size of 375 x 42, with the colors given as parameters and without an element that is shown directly.
     * 
     * @param elements
     *      A list of elements you want to add in the list.
     *      If you want an empty list you have to give an empty java.util.List<Element> as parameter but not null.
     *      Null is not excepted as a parameter.
     * 
     * @param listColor
     *      The background color of the list.
     * 
     * @param textColor
     *      The color of the text in the list.
     * 
     * @throws
     *      IllegalArgumentException is thrown if the parameter elements is null or if the Color listColor is null.
     */
    public DropdownList(List<? extends Element> elements, Color listColor, Color textColor) throws IllegalArgumentException {
        this(elements, null, 375, 40, listColor, textColor);
    }
    /**
     * Creates a new DropdownList with the given elements, a default size of 375 x 42, with the colors given as parameters ground and with an element that is shown directly.
     * 
     * @param elements
     *      A list of elements you want to add in the list.
     *      If you want an empty list you have to give an empty java.util.List<Element> as parameter but not null.
     *      Null is not excepted as a parameter.
     * 
     * @param firstShownElement
     *      The element that is shown directly after creating the DropdownList.
     * 
     * @param listColor
     *      The background color of the list.
     * 
     * @param textColor
     *      The color of the text in the list.
     * 
     * @throws
     *      IllegalArgumentException is thrown if the parameter elements is null or if the Color listColor is null.
     */
    public DropdownList(List<? extends Element> elements, Element firstShownElement, Color listColor, Color textColor) throws IllegalArgumentException {
        this(elements, firstShownElement, 375, 40, listColor, textColor);
    }
    /**
     * Creates a new DropdownList with the given elements, a default size as given in the parameters, with black letters on white ground and without an element that is shown directly.
     * 
     * @param elements
     *      A list of elements you want to add in the list.
     *      If you want an empty list you have to give an empty java.util.List<Element> as parameter but not null.
     *      Null is not excepted as a parameter.
     * 
     * @param width
     *      The width of the list (has to be greater of equal 50).
     * 
     * @param height
     *      The hight of the list (has to be greater or equal 20).
     * 
     * @throws
     *      IllegalArgumentException is thrown if the parameter elements is null or if the parameters for width or height arent possible.
     */
    public DropdownList(List<? extends Element> elements, int width, int height) throws IllegalArgumentException {
        this(elements, null, width, height, Color.white, Color.black);
    }
    /**
     * Creates a new DropdownList with the given elements, a default size of 375 x 42, with black letters on white ground and with an element that is shown directly.
     * 
     * @param elements
     *      A list of elements you want to add in the list.
     *      If you want an empty list you have to give an empty java.util.List<Element> as parameter but not null.
     *      Null is not excepted as a parameter.
     * 
     * @param firstShownElement
     *      The element that is shown directly after creating the DropdownList.
     * 
     * @param width
     *      The width of the list (has to be greater of equal 50).
     * 
     * @param height
     *      The hight of the list (has to be greater or equal 20).
     * 
     * @throws
     *      IllegalArgumentException is thrown if the parameter elements is null or if the parameters for width or height arent possible.
     */
    public DropdownList(List<? extends Element> elements, Element firstShownElement, int width, int height) throws IllegalArgumentException {
        this(elements, firstShownElement, width, height, Color.white, Color.black);
    }
    /**
     * Creates a new DropdownList with the given elements, a default size of 375 x 42, with black letters on white ground and without an element that is shown directly.
     * 
     * @param elements
     *      A list of elements you want to add in the list.
     *      If you want an empty list you have to give an empty java.util.List<Element> as parameter but not null.
     *      Null is not excepted as a parameter.
     * 
     * @param width
     *      The width of the list (has to be greater of equal 50).
     * 
     * @param height
     *      The hight of the list (has to be greater or equal 20).
     * 
     * @param listColor
     *      The background color of the list.
     * 
     * @throws
     *      IllegalArgumentException is thrown if the parameter elements is null or if the parameters for width or height arent possible or if the Color listColor is null.
     */
    public DropdownList(List<? extends Element> elements, int width, int height, Color listColor) throws IllegalArgumentException {
        this(elements, null, width, height, listColor, Color.black);
    }
    /**
     * Creates a new DropdownList with the given elements, a default size of 375 x 42, with black letters on white ground and with an element that is shown directly.
     * 
     * @param elements
     *      A list of elements you want to add in the list.
     *      If you want an empty list you have to give an empty java.util.List<Element> as parameter but not null.
     *      Null is not excepted as a parameter.
     * 
     * @param firstShownElement
     *      The element that is shown directly after creating the DropdownList.
     * 
     * @param width
     *      The width of the list (has to be greater of equal 50).
     * 
     * @param height
     *      The hight of the list (has to be greater or equal 20).
     * 
     * @param listColor
     *      The background color of the list.
     * 
     * @throws
     *      IllegalArgumentException is thrown if the parameter elements is null or if the parameters for width or height arent possible or if the Color listColor is null.
     */
    public DropdownList(List<? extends Element> elements, Element firstShownElement, int width, int height, Color listColor) throws IllegalArgumentException {
        this(elements, firstShownElement, width, height, listColor, Color.black);
    }
    /**
     * Creates a new DropdownList with the given elements, a default size of 375 x 42, with black letters on white ground and without an element that is shown directly.
     * 
     * @param elements
     *      A list of elements you want to add in the list.
     *      If you want an empty list you have to give an empty java.util.List<Element> as parameter but not null.
     *      Null is not excepted as a parameter.
     * 
     * @param width
     *      The width of the list (has to be greater of equal 50).
     * 
     * @param height
     *      The hight of the list (has to be greater or equal 20).
     * 
     * @param listColor
     *      The background color of the list.
     * 
     * @param textColor
     *      The color of the text in the list.
     * 
     * @throws
     *      IllegalArgumentException is thrown if the parameter elements is null or if the parameters for width or height arent possible or if the Color listColor is null.
     */
    public DropdownList(List<? extends Element> elements, int width, int height, Color listColor, Color textColor) throws IllegalArgumentException {
        this(elements, null, width, height, listColor, textColor);
    }
    /**
     * Creates a new DropdownList with the given elements, a default size of 375 x 42, with black letters on white ground and with an element that is shown directly.
     * 
     * @param elements
     *      A list of elements you want to add in the list.
     *      If you want an empty list you have to give an empty java.util.List<Element> as parameter but not null.
     *      Null is not excepted as a parameter.
     * 
     * @param firstShownElement
     *      The element that is shown directly after creating the DropdownList.
     * 
     * @param width
     *      The width of the list (has to be greater of equal 50).
     * 
     * @param height
     *      The hight of the list (has to be greater or equal 20).
     * 
     * @param listColor
     *      The background color of the list.
     * 
     * @param textColor
     *      The color of the text in the list.
     * 
     * @throws
     *      IllegalArgumentException is thrown if the parameter elements is null or if the parameters for width or height arent possible or if the Color listColor is null.
     */
    public DropdownList(List<? extends Element> elements, Element firstShownElement, int width, int height, Color listColor, Color textColor) throws IllegalArgumentException {
        if (width < 50) {
            throw new IllegalArgumentException("The width(" + width + ") of the dropdownlist has to be greater or equal 50.");
        }
        else if (height < 20) {
            throw new IllegalArgumentException("The height(" + height + ") of the dropdownlist has to be greater or equal 20.");
        }
        if (elements == null) {
            throw new IllegalArgumentException("List<Element> elements mussn#t be null.");
        }
        if (listColor == null || textColor == null) {
            throw new IllegalArgumentException("The colors listColor and textColor mussn't be null.");
        }
        listElements = new ArrayList<DropdownList>();
        this.width = width;
        this.height = height;
        this.listColor = listColor;
        this.textColor = textColor;
        for (Element element : elements) {
            listElements.add(new DropdownList(element, this, width, height, listColor, textColor));
        }
        elementOfList = false;
        getImage().clear();
        getImage().scale(width, height);
        getImage().setColor(Color.black);
        getImage().fill();
        getImage().setColor(listColor);
        getImage().fillRect(1, 1, width - 2, height - 2);
        if (firstShownElement != null) {
            setElement(firstShownElement);
        }
        GreenfootImage arrow = new GreenfootImage("arrow.png");
        arrow.scale((int) (((double) (height - 4) / (double) arrow.getHeight()) * arrow.getWidth()), (height - 4));
        getImage().drawImage(arrow, getImage().getWidth() - arrow.getWidth() - 2, getImage().getHeight()/2 - arrow.getHeight()/2);
    }
    
    /**
     * Constructs a Element of the DropdownList.
     * 
     * @param element
     *      The Element that this listpart should show.
     * 
     * @param callingList
     *      A reference to the calling list to send the values to it.
     * 
     * @param width
     *      The width of the element which is the same than the width of the calling list.
     * 
     * @param height
     *      The height of the element which is the same than the height of the calling list.
     * 
     * @param listColor
     *      The background color of the element which is the same than the background color of the calling list.
     * 
     * @param textColor
     *      The text color of the element which is the same than the text color of the calling list.
     */
    private DropdownList(Element element, DropdownList callingList, int width, int height, Color listColor, Color textColor) {
        this.element = element;
        this.callingList = callingList;
        this.width = width;
        this.height = height;
        this.listColor = listColor;
        this.textColor = textColor;
        elementOfList = true;
        getImage().clear();
        getImage().scale(width, height);
        getImage().setColor(Color.black);
        getImage().fill();
        getImage().setColor(listColor);
        getImage().fillRect(1, 1, width - 2, height - 2);
        GreenfootImage img = new GreenfootImage(element.getName(), height - 5, textColor, new Color(0, 0, 0, 0));
        getImage().drawImage(img, 5, getImage().getHeight()/2 - img.getHeight()/2);
    }
    
    /**
     * Select a element of the list and open or close the list.
     */
    public void act() {
        if (elementOfList) {
            if (Greenfoot.mouseClicked(this)) {
                if (!this.element.getName().equals("<no elements>")) {
                    callingList.resetImage();
                    callingList.setElement(this.element);
                    fireActionEvent(GreenfootActionEvent.DROPDOWN_LIST_SELECTED_ELEMENT);
                }
                callingList.removeElements();
            }
        }
        else {
            if (Greenfoot.mouseClicked(this)) {
                if (listDisplayed) {
                    removeElements();
                }
                else {
                    if (listElements.isEmpty()) {
                        Element element = new Default_Element("<no elements>");
                        emptyElement = new DropdownList(element, this, width, height, listColor, textColor);
                        GreenfootImage text = new GreenfootImage(element.getName(), height - 5, textColor, new Color(0, 0, 0, 0));
                        emptyElement.getImage().drawImage(text, 5, getImage().getHeight()/2 - text.getHeight()/2);
                        getWorld().addObject(emptyElement, getX(), getY() + (getImage().getHeight() / 2) + (emptyElement.getImage().getHeight() / 2) - 1);
                    }
                    else {
                        for (DropdownList element : listElements) {
                            getWorld().addObject(element, getX(), getY() + (getImage().getHeight() / 2) + (element.getImage().getHeight() / 2) + (height * listElements.indexOf(element)) - 1);
                        }
                    }
                    listDisplayed = true;
                }
            }
        }
    }
    
    /**
     * Fire an GreenfootActionEvent with the given action number to the current GreenfootActionListener.
     * 
     * @param action
     *      The event's number.
     */
    private void fireActionEvent(int action) {
        if (greenfootActionListener != null) {
            greenfootActionListener.actionPerformed(new GreenfootActionEvent(action));
        }
        else if (callingList.getActionListener() != null) {
            callingList.getActionListener().actionPerformed(new GreenfootActionEvent(action));
        }
    }
    
    /**
     * Set the current element of a DropdownList to the given element.
     * 
     * @param element
     *      The element that is chosen in the list.
     */
    public void setElement(Element element) {
        this.element = element;
        GreenfootImage text = new GreenfootImage(element.getName(), height - 5, textColor, new Color(0, 0, 0, 0));
        getImage().drawImage(text, 7, getImage().getHeight()/2 - text.getHeight()/2);
    }
    
    /**
     * Remove the whole list and all elements.
     */
    public void remove() {
        for (DropdownList element : listElements) {
            getWorld().removeObject(element);
        }
        getWorld().removeObject(emptyElement);
        getWorld().removeObject(this);
    }
    /**
     * Removes all elements of the list (closes the list).
     */
    public void removeElements() {
        for (DropdownList element : listElements) {
            try {
                getWorld().removeObject(element);
            }
            catch (NullPointerException npe) {
                //npe.printStackTrace();
            }
        }
        try {
            getWorld().removeObject(emptyElement);
        }
        catch (NullPointerException npe) {
            //npe.printStackTrace();
        }
        listDisplayed = false;
    }
    
    /**
     * Resets the image of the dropdown_List.
     */
    private void resetImage() {
        getImage().clear();
        getImage().scale(width, height);
        getImage().setColor(Color.black);
        getImage().fill();
        getImage().setColor(listColor);
        getImage().fillRect(1, 1, width - 2, height - 2);
        GreenfootImage arrow = new GreenfootImage("arrow.png");
        arrow.scale((int) (((double) (height - 4) / (double) arrow.getHeight()) * arrow.getWidth()), (height - 4));
        getImage().drawImage(arrow, getImage().getWidth() - arrow.getWidth() - 2, getImage().getHeight()/2 - arrow.getHeight()/2);
    }
    
    /**
     * Returns the currently selected element.
     * 
     * @return
     *      The currently selected element that is shown in the list.
     */
    public Element getElement() {
        return element;
    }
    
    /**
     * Get the current GreenfootActionListener.
     * 
     * @return
     *      The current GreenfootActionListener.
     */
    public GreenfootActionListener getActionListener() {
        return greenfootActionListener;
    }
    /**
     * Change the GreenfootActionListener of this object.
     * 
     * @param greenfootActionListener
     *      The new GreenfootActionListener for this object.
     */
    public void addGreenfootActionListener(GreenfootActionListener greenfootActionListener) {
        this.greenfootActionListener = greenfootActionListener;
    }
    
    /**
     * This class is just a default element for the dropdownlist which is added if there is no other element.
     * The only important thing in this class is that it has to implement Element to be a element of the list.
     * 
     * Because it implements Element (see the class declaration) it has to overwrite the method getName() in Element and can be casted to Element.
     * 
     * If you want to add other classes to the dropdownlist this classes have to implement Element to and have to overwrite the getName() method.
     * In this method you can return whatever you want. It doesn't need to be the String name.
     * 
     */
    class Default_Element extends Actor implements Element
    {
        String name;
        
        public Default_Element(String name) {
            this.name = name;
        }
        
        public String getName() {
            return name;
        }
    }
}